/*
 * Convert MPU data for SPARCstation /dev/audio
 * Permit real-time adjustment of playing parameters
 * SAU 3/93
 *
 * Strategy:
 *  Instead of using the obvious strategy of synthesizing each note as the
 *  key goes up, we wait until we need to output a sample for that note.
 *  This permits more rapid changes in
 *  the performance parameters.  As we expect this program to be called
 *  from some interactive user interface, at each call we check the current
 *  performance characteristics, and output a HUNK, reading midi events only
 *  as necessary. If we are outputting to /dev/audio (the default) we control
 *  the size of the kernel buffers so performance changes happen quickly.
 */

/* debugging flags
 * c  each midi command
 * e  error condition
 * f  time adjusting
 * l  list operations
 * m  main loop
 * n  note rendering/deleting
 * o  output
 * p  Command processing
 */

#include <fcntl.h>
#include <sys/time.h>
#include <stdio.h>
#include <stropts.h>
#include <sys/select.h>
#include	"libmidi.h"
#include	"libmpu.h"
#include "synth.h"
#include "devaudio.h"

#define GET_OPT(i)   \
   strlen(argv[i])>2 ? argv[i]+2 : argv[++i]

char debug_list[20];	/* list of debugging stuff */

char *sine_table();
void do_sines();
int sine_resp();
int to_audio();

static void set_params();
static int wait_read();
static void term();

/* these are the changable performance parameters; pick some defaults */

struct params param_data = {
	NULL,		/* file */
	1000,		/* gain factor (% * 10) */
	100,		/* tempo (bpm) */
	100,		/* OLD tempo (bpm) */
	0,			/* transposition (in semi-tones) */
	110,		/* note duration (%) */ 
	4,			/* velocity factor (0-10) */
	0,			/* which tuning: 0=equal, 1=true */
	4000,		/* bytes spit-out at each iteration */
	0,			/* generate verbose information */
	0,			/* state */
	0,			/* channels used */
	0x0,		/* balance (1=left, 2=right, 0 or 3 = both */
	0xFF,		/* channel mask (start with 7 channels */
	0,			/* which voice to change */
	16000,	/* sampling rate */
	2,			/* number of channels */
	{5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5},	/* individual channel gains */
	NULL		/* shapes */
	};

struct part weights[10] = {
	{0, 50, 1},
	{1, 40, 2}, 
	{2, 30, 3},
	{3, 30, 4},
	{4, 30, 5},
	{5, 30, 6},
	{0,  0, 0},
	};
	
struct voice builtin_voice = {
	0,			/* voice number */
	7000,		/* cutoff freq */
	1,			/* stretch */
	weights,	/* partial weights */
	NULL,		/* no shape for now */
	};

char line[MAXLINE];	/* stdin/stdout buffer */
data buff[MAXBUFF];	/* where the audio goes */
short out[2*MAXBUFF];	/* where the converted output goes */
struct voice voices[16];	/* parameters for the 16 midi voices */

extern char *malloc();
extern char *getenv();

/* list management functions */

static void print_list(/* struct note *list */);
char *parse_command(/* params,voice,line */);

/* other stuff */

int play(/* int *buff, char *out, int n,int fd, int gain */);
int play_mu(/* int *buff, char *out, int n,int fd, int gain */);
int audio_type(/* int fd */);
struct audio_state *init_audio(/* struct audio_state *audio */);

/* main program */

char *table[MIDI_MAX_CHANS];	/* pointers to partial tables */

int
main(argc, argv)
int argc;
char *argv[];
	{
	int i;			/* index */
	int q;			/* bytes above high water mark */
	int n;			/* bytes of audio processed */
	char *audio_dev;	/* audio cookie */
	int xtra = 8000;	/* extra bytes in audio Q */
	FILE *file=NULL;	/* config file fd */
	int no_audio=0;	/* true if output is NOT an audio device */
	int mono=0;			/* mono only */
	int rate=0;			/* rate from command line */
	char **files = NULL;	/* list of initial files to play */

	char *getenv();
	struct params *params = &param_data;	/* performance data */
	struct audio_state *audio;					/* audio progress data */

	if (getenv("DEBUG")) {
		strcpy(debug_list,getenv("DEBUG"));
		fprintf(stderr,"Debug list: %s\n",debug_list);
		}

	if (getenv("ERR_FILE")) {
		freopen(getenv("ERR_FILE"),"w",stderr);
		}

	if (getenv("EXTRA")) 
		xtra = atoi(getenv("EXTRA"));

	/* process arguments */

   for(i=1;i<argc;i++) {
      if (*argv[i] == '-')
         switch (argv[i][1]) {
            case 'm':            /* use mono mu-law mode */
               mono++;
               break;
				case 'r':				/* set rate */
					rate = atoi(GET_OPT(i));
					break;
				case 'c':				/* get config file */
					file = fopen(GET_OPT(i),"r");
					P(state) = 0;
					break;
				case 'x':				/* get xtra samples */
					xtra = atoi(GET_OPT(i));
					break;
				default:
					fprintf(stderr,"Invalid flag, %s, ignored\n",argv[i]);
				} 
			else if (!files) {		/* set files to rest of args */
				files = &argv[i];
			}
		}
		
	/* initialize the audio */

	audio = init_audio(NULL);
	audio_dev = audio_open();	
	if (audio_dev == NULL) {
		fprintf(stderr,"Audio device won't open, sorry\n");
		term("audio is busy, try later",1);
		}
	no_audio = ((audio_inq(audio_dev) == -1));

	/* get the hardware audio modes */

	if (!no_audio)  {
			int mode = audio_modes(audio_dev);
			if (mono) mode &= ~2;
			switch(mode) {
			default:		/* no audio at all! */
				fprintf(stderr,"Can't find audio, sorry\n");
				exit(1);
			case 1:		/* AMD codec (sparc 1's and 2's) */
				param_data.stereo = 1;		/* 1 channel, mu-law */
				param_data.rate = 8000;	/* only supported rate */
				builtin_voice.cutoff = 3000;
				weights[4].weight = 0;	/* this is only needed for sparc 1's */
				xtra = 2000;
				break;
			case 2: case 3:		/* Speaker box - the default */
				if (rate && rate_ok(rate))
					param_data.rate = rate;
				else if (rate)
					fprintf(stderr,"Invalid rate, %d, using default\n",rate);
				break;
			}
		}
	else if (mono) {		/* mono, no audio */
		param_data.stereo = 1;		/* 1 channel, mu-law */
		param_data.rate = 8000;	/* only supported rate */
		builtin_voice.cutoff = 4000;
		xtra = 2000;
		audio_rates = NULL;		/* kludge to permit ANY rate */
		}

	/* close the audio, reopen when needed */

	if (!no_audio) {
		audio_close(audio_dev);
		audio_dev = NULL;
		}

	/* initialize the voices (this is temporary - to alloc space) */

	for(i=0;i<MIDI_MAX_CHANS;i++) {
		int size = 10 * sizeof(struct part);
		voices[i] = builtin_voice;
		voices[i].weights = (struct part *) malloc(size);
		bcopy(weights,voices[i].weights,size);
		voices[i].id = i;
		}

	/* read voice init file */

	if (file) {
		dprintf('m')(stderr,"Reading parameter file\n");
		while(fgets(line,MAXLINE,file)) {
			parse_command(params,voices,line);
			}
		fclose(file);
		}

	for(i=0;i<MIDI_MAX_CHANS;i++)
		table[i] = sine_table(NULL,&voices[i],P(tuning),(data) P(rate), (data) 1, (data) 1);

	while(1) {

		/* process command line files */

		if (!P(file) && files) {		/* playing files from command line */
			while(*files && (P(file) = fopen(*files++,"r"))==NULL);
			if (P(file)) {
				dprintf('M')(stderr,"Started file %s\n",*argv);
				P(state) = RESTART|RUNNING;
				}
			else {							/* done with command line files, exit */
				term("ran out of files",0);
				}
			}

		if (P(state)&RUNNING && !no_audio) {
         if (audio_dev == NULL) {    /* reopen audio (if needed) */
            dprintf('m')(stderr,"Opening audio\n");
            audio_dev = audio_open();
				audio_setmode(audio_dev,(int) P(rate),P(stereo),P(stereo));
            }
			q = audio_inq(audio_dev)-(P(hunk)+xtra);
			}
		else
			q = P(hunk);

		if (q > 0 || no_audio) {
			set_params(params,voices,A(id),line,MAXLINE,q);
			}

		if (q <= 0 || no_audio) {
			if (P(state)&(NEW_VOICE|RESPONSE)) {		/* do a voice calculation */
				int i = 0;
				dprintf('v')(stderr,"new voice 0x%x\n",P(voice));
				for(i=0;P(voice);P(voice)>>=1,i++) {
					if (P(voice)&1) {
						dprintf('v')(stderr," new table for voice %d\n",i);
						table[i] = sine_table(table[i],voices+i,P(tuning),
								(data) P(rate),(data) 1, (data) 1);
						/* sine_resp(table[i],P(shape),(data) P(rate)); /* wrong */
						}
					}
				P(state) &= ~(NEW_VOICE|RESPONSE);
				}
			n = to_audio(audio,params,buff,(data) P(rate));
			if (P(stereo)==2)
				play(buff,out,2*n,audio_dev,P(gain));
			else
				play_mu(buff,out,n,audio_dev,P(gain));
			}

      /* see if we need to close the audio device */

      if (!no_audio && (P(state)&RUNNING)==0 && audio_dev) {
         dprintf('m')(stderr,"Closing audio\n");
         audio_close(audio_dev);
         audio_dev = NULL;
         }
		}
	return(0);
	}

/* convert to mu-law, then write */

#define B13ZERO 0x1000		/* zero offset in unsigned 13-bit tables */  
extern unsigned char Ub2i[];

int
play_mu(buff,out,n,fd,gain)
data *buff;    /* buffer of sound data */
short *out;     /* the output mu-law buffer */
int n;         /* number of samples */
char *fd;		/* fd to write output to */
int gain;      /* % of nominal gain * 10 */
   {
   char *p = (char *) out;
   int count = n;
   int clip = 0;
   int max = 0;
   unsigned char *cnvt;
   register int sample;
   cnvt = Ub2i+B13ZERO;

   while(n-- > 0) {
      sample = gain * *buff++ / 3000;
#ifndef NOCLIP
      if (sample < -4010 || sample > 4010)  {   /* for testing */
         if (abs(sample) > max) max = abs(sample);
         sample = 0; /* wrong */
         clip++;
         }
#endif
      *p++ = cnvt[sample];
      }
   if (clip)
      fprintf(stderr,"clip %d/%d, max value = %d\n",clip,count,max);
	return(audio_write(fd,(char *)out,count));
   }

/* scale then output as linear */

int
play(buff,out,n,fd,gain)
data *buff;		/* buffer of sound data */
short *out;		/* the output mu-law buffer */
int n;			/* number of samples */
char *fd;			/* file descriptor to output to */
int gain;		/* % of nominal gain * 10 */
	{
	data *end = buff + n;
	data mult = gain/1000.0;

	while(buff < end) {
		*out++ = mult * *buff++;
		}

	return(audio_write(fd,(char *) (out-n),n*sizeof(*out)));
	}

/* read in a command line, for interactive use */

static void
set_params(params,voice,id,line,max,wait)
struct params *params;	/* current performance parameters */
struct voice voice[];		/* set of voice parameters */
int id;			/* current note id (for status) */
char *line;
int max;			/* max chars to wait */
int wait;		/* number of samples to wait before reading */
	{
	char *msg;		/* message returned from parse_command */
	int n;

	if ((n=wait_read(0,line,max,wait))>0) {
		if (n>1) dprintf('p')(stderr,"Got command: %s\n",line);
		msg = parse_command(params,voice,line);
		printf("* %d %c %.04x {%s}\n",
				id,P(state)&RUNNING?'+':'-',P(channels),msg);
		fflush(stdout);
		P(channels) = 0;
		}
	}

/* sum a hunk.  render notes as required */
/* this function belongs in to_audio.c */

int
output(audio,params,buff,channels)
struct audio_state *audio;
struct params *params;	/* rendering parameters */
data *buff;					/* where to put the output */
int channels;				/* either 1 or 2 */
	{
	struct note *list = A(list);
	int done = A(done);
	int count = 0;			/* number of completed notes seen */
	int delta;				/* # samples from end of note */
	struct note *note;	/* current note */
	data *start_note;		/* starting sample to render in note */
	int n;					/* # of samples from this note */
	data *p;					/* starting sample to render in buff */

	dprintf('o')(stderr,"outing hunk %d-%d\n",done,done+P(hunk));
	bzero(buff,channels*P(hunk)*sizeof(data));
	for(note=list;note;note=note->next) {

		/* skip notes not participating in this hunk */

		if (note->stop < done){					/* done with this puppy */
			dprintf('o')(stderr,"   %d (%d:%d) is done (%ld->%ld < %d)\n",
				note->id,note->chan,note->key,note->start,note->stop,done);
			count++;
			continue;
			}
		if (note->start >= done + P(hunk)) {
			dprintf('o')(stderr,"   %d (%d:%d) hasn't started (%ld-%ld > %d)\n",
				note->id,note->chan,note->key,note->start,note->stop,done);
			continue;
			}
		if (note->start > note->stop) {
			fprintf(stderr,"AARGH!, %d (%d:%d) %ld->??, nothing to render\n",
				note->id,note->chan,note->key,note->start);
			print_list(list,done,A(current));
			continue;
			}

		/* skip notes for invalid channel mask */

		if (!((1<<note->chan) & P(mask)))
			continue;

		/* render the note */

		if (!note->data && note->stop>note->start) {
			int dur = (note->stop-note->start) * P(duration)/100;

			dur &= ~1;		/* need even # of samples */
			note->stop = note->start + dur-1;	/* off by 1? */
			note->data = (data *) malloc(dur*sizeof(data));
			dprintf('n') (stderr," rendering %d (%d:%d) %ld->%ld (%d) at 0x%x\n",
					note->id,note->chan,note->key,note->start,
					note->stop,dur,(unsigned int)note->data);
			do_sines(table[note->chan],note->key+P(trans),dur,
						note->vel,P(gains)[note->chan],P(factor),note->data);
			}

		/* compute boundary conditions */

		if ((delta = note->start-done) > 0) {	/* note > hunk */
			start_note = note->data;
			p = buff + channels*delta;
			n = P(hunk) - delta;
			dprintf('o')(stderr," %d (%d:%d) adjusting start by %d, (%d bytes)\n",
					note->id, note->chan,note->key,delta,n);
			}
		else {							/* note starts before beginning of hunk */
			start_note = note->data - delta;
			p = buff;
			n = P(hunk);
			}

		if ((delta = note->stop - (done+P(hunk))) < 0) {	/* note < hunk */
			n += delta;
			dprintf('o')(stderr," %d (%d:%d) adjusting end by %d, (%d bytes)\n",
					note->id, note->chan,note->key,delta,n);
			}

		if (n<0) {
			fprintf(stderr,"done=%d, hunk=%d, note len=%d\n",
				done,P(hunk),n);
			fprintf(stderr,"AARG! note %d is too short!\n",note->id);
			print_list(list,done,A(current));
			continue;
			}
			
		/* add note to sound */
		
		if (channels == 1)
			while (n-->0) 
				*p++ += *start_note++;
		else {
			switch(BALANCE(note->chan)) {
				case RIGHT:	
					p++; /* no break */
				case LEFT:
					while (n-->0) {
						*p += 2.0 * *start_note++;
						p += 2;
						}
					break;
				default:
					while (n-->0) {
						*p++ += *start_note;
						*p++ += *start_note++;
						}
					break;
				}
			}
		}

	return(count);
	}

/* read from fd non blocking */

static int
wait_read(fd,buff,n,samples)
int fd;			/* file descriptor to read from */
char *buff;		/* read into here */
int n;			/* max # to read */
int samples;	/* # of samples to wait */
	{
	fd_set mask;
	int cnt=0;
   struct timeval time;
   time.tv_sec = (samples)/8000;
   time.tv_usec = ((samples)%8000)*1000/8;
	FD_ZERO(&mask);
	FD_SET(fd,&mask);

	if (select(32,&mask,0,0,&time)==1) {	/* read from stdin (if avaliable) */
		cnt = read(fd, buff, n);
		}
	return(cnt);
	}

/* print a note list (for debugging) */

static void
print_list(list,done,current)
struct note *list;
int done, current;
	{
	fprintf(stderr,"note list: dont=%d, current=%d\n",done,current);
	for(;list;list=list->next) {
		fprintf(stderr,"%d) %d:%d (%ld->%ld) %s\n",
			list->id, list->chan, list->key,list->start,list->stop,
			list->data ? "rendered":"");	
		}
	}

/* die, but let user interface know first */

void
term(s,n)
char *s;
int n;
	{
	fprintf(stderr,"Synth aborting: %s\n",s);
	alarm(5);
	gets(line);
	alarm(0);
	printf("Xit: %s\n",s);
	exit(n);
	}

/* see if sampling rate is OK 
 * If np rates, then ANY rate is OK OK?
 */

int
rate_ok(n)
int n;
	{
	int i;
	if (audio_rates == NULL) return (1);
	for(i=0;audio_rates[i];i++)
		if (audio_rates[i] == n)
			return(1);
	return(0);
	}
